from pydantic import BaseModel, Field
from mcp.server.fastmcp.tools import ToolManager
import mcp

from typing import Any, List

from .types.backend import Backend
from .types.message import Message, MessageComponent, MessageToolResultComponent, MessageToolCallComponent
from .config import Config

import json

import logging
logger = logging.getLogger(__name__)

class Agent(BaseModel):
    config: Config = Field(default_factory=Config)
    backend: Backend

    __tool_manager: ToolManager

    def model_post_init(self, ctx):
        self.__tool_manager = ToolManager()

    @property
    def add_tool(self): 
        return self.__tool_manager.add_tool

    @property
    def list_tools(self): 
        return self.__tool_manager.list_tools

    def _call_tool(self, tool_name: str, *args, **kvargs) -> Any:
        tool = self.__tool_manager.get_tool(tool_name)

        if not tool:
            logger.debug(f"Agent requested unknown tool {tool_name}")
            return f"ERROR: Unknown tool: '{tool_name}'"
        
        try:
            tool_result = tool.fn(*args, **kvargs)
        except Exception as err:
            logger.error(f"Exception occured while calling {tool_name}: {err}")
            return f"ERROR: Exception occured while calling {tool_name}: {err}"

        return tool_result

    def _workaround_system_prompt(self, tools: List[mcp.Tool]) -> str:
        system_prompt = self.config.system_prompt

        system_prompt += """# TOOL CALLING
TOOL CALLS ARE PERFORMED IN JSON OBJECTS THAT LOOK AS FOLLOWS:
{{
    "action": "tool_call",
    "tool_call": {{
        "tool_name": "<tool name>",
        "input": <input object/data>
    }}
}}
Responses that contain anything more WILL BE REJECTED.


# TOOLS
"""
        for tool in tools:
            system_prompt += f"""
tool_name: {tool.name}
{tool.description}

input_schema: 
{json.dumps(tool.parameters)}

outputs:
{tool.fn_metadata.output_schema}"""
        return system_prompt

    def _workaround_map_message_to_tool_call(self, workaround_message: Message) -> Message:
        # Last message should always be a content block
        workaround_message_component = workaround_message.components[-1]

        if workaround_message_component.type != 'message':
            return workaround_message

        try:
            # Attempt to extract a json object
            json_content = workaround_message_component.content.removeprefix('```').removeprefix('json').removesuffix('```')

            tool_call_obj = json.loads(json_content)
        
            return Message(
                role = 'assistant',
                components  = [
                    MessageToolCallComponent(
                        type='tool_call',
                        content=mcp.types.CallToolRequest(
                            method = 'tools/call',
                            params = mcp.types.CallToolRequestParams(
                                name = tool_call_obj['tool_call']['tool_name'],
                                arguments = tool_call_obj['tool_call']['input']
                            )
                        )
                    )
                ]
            )
        except Exception as _:
            return workaround_message

    

    def _invoke_completion(self, messages: List[Message], stop_on: List[str], force_tools: bool = True) -> List[Message]:
        if len(stop_on) == 0 and force_tools:
            raise Exception(f"Empty `stop_on` and `force_tools` = True not supported.")

        tools = self.list_tools()

        # PREPARE LLM RESPONSE ARGS
        _kvargs = {
            'agent_config': self.config,
        }

        while True:
            # PREPARE LLM RESPONSE ARGS
            _kvargs['messages'] = messages

            if self.config.do_unsupported_model_workaround:
                _kvargs['system_prompt_override'] = self._workaround_system_prompt(tools)

            else:
                _kvargs['tools'] = tools
                
            # GENERATE RESPONSE
            logger.debug(f"Generating response from backend({self.backend.__class__.__qualname__}): {_kvargs}")
            response =  self.backend.generate_response(**_kvargs)
            messages.append(response)
            logger.debug(f"Received response from backend({self.backend.__class__.__qualname__}): {response}")

            # RETURN IF NO TOOL USE REQUIRED
            if not force_tools:
                return messages

            # HANDLE WORKAROUND
            if force_tools and self.config.do_unsupported_model_workaround:
                response = self._workaround_map_message_to_tool_call(response)


            # TOOL USE ENFORCEMENT
            if force_tools and response.components[-1].type != 'tool_call':
                logger.debug(f"`force_tools` enabled and no tool response received. Re-prompting....")
                messages.append(Message(role='user', components=[MessageComponent(type='message', content="ERROR: Calling a tool is REQUIRED")]))
                continue

            # EXECUTE TOOLS
            for tool_call in filter(lambda x: x.type == 'tool_call', response.components):
                tool_call_id = tool_call.meta.get('tool_call_id')

                logger.debug(f"Invoking tool_call({tool_call_id}) ({tool_call.content.params})")
                result = self._call_tool(tool_call.content.params.name, **tool_call.content.params.arguments)
                logger.debug(f"Result of tool_call({tool_call_id}) ({result})")

                messages.append(Message(
                    role='system',
                    components=[
                        MessageToolResultComponent(
                            meta={
                                'tool_call_id': tool_call_id, 
                                'tool_call_name': tool_call.content.params.name
                            },
                            type='tool_result',
                            content=mcp.types.CallToolResult(
                                content=[mcp.types.TextContent(type='text', text=str(result))]
                            ),
                        )
                    ]
                ))


                # TOOL STOP
                if tool_call.content.params.name in stop_on:
                    logger.debug(f"Stopping after tool_call({tool_call_id}) ({tool_call.content.params})")
                    return messages

    def invoke(self, user_message: str, stop_on: List[str] = [], force_tools: bool = True) -> List[Message]:
        messages = [
            Message(
                role='user',
                components=[MessageComponent(
                    type='message',
                    content=user_message
                )],
            )
        ]

        return self._invoke_completion(messages=messages, stop_on=stop_on, force_tools=force_tools)


# Alias
Agnet = Agent


from pydantic import BaseModel, Field

from ..config import Config
from .message import Message

from typing import Dict, List
import abc

import mcp.server.fastmcp.tools

class Backend(BaseModel, abc.ABC):
    config: Dict[str,str] = Field(default_factory=dict)

    @abc.abstractmethod
    def generate_response(self, messages: List[Message], agent_config: Config, tools: List[mcp.server.fastmcp.tools.Tool] = [], system_prompt_override: str = "", **kvargs) -> Message:
        pass



from pydantic import BaseModel, Field


class Config(BaseModel):
    model_name: str
    system_prompt: str = Field(default="")

    
    do_thinking: bool = Field(default=False)
    do_unsupported_model_workaround: bool = Field(default=False)


from pydantic import BaseModel, Field
import mcp.types

from typing import Literal, List, Dict, Any


class MessageComponent(BaseModel):
    type: Literal['message']
    content: str
    meta: Dict[str, Any] = Field(default_factory=dict)

class MessageThinkingComponent(MessageComponent):
    type: Literal['thinking']
    content: str

class MessageToolCallComponent(MessageComponent):
    type: Literal['tool_call']
    content: mcp.types.CallToolRequest

class MessageToolResultComponent(MessageComponent):
    type: Literal['tool_result']
    content: mcp.types.CallToolResult



class Message(BaseModel):
    role: Literal['system', 'assistant', 'user']
    components: List[MessageComponent | MessageThinkingComponent | MessageToolCallComponent | MessageToolResultComponent]


from openai import OpenAI
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.chat_completion_tool_param import ChatCompletionToolParam
from openai.types.shared_params.function_definition import FunctionDefinition

from ..types.backend import Backend
from ..types.message import Message, MessageComponent, MessageThinkingComponent, MessageToolCallComponent

from typing import Dict

import mcp.server.fastmcp.tools

import json

def _map_to_openai_message(message: Message) -> Dict:
    msg = {
        "role": message.role,
    }

    for component in message.components:
        if component.type == 'message':
            msg['content'] = component.content
            continue

        if component.type == 'thinking':
            msg['thinking'] = component.content
            continue

        if component.type == 'tool_call':
            if not msg.get('tool_calls'):
                msg['tool_calls'] = []

            msg['tool_calls'].append({
                'id': component.meta.get('tool_call_id'),
                'type': 'function',
                'function': {
                    'name': component.content.params.name,
                    'arguments': json.dumps(component.content.params.arguments)
                }
            })
            continue

        if component.type == 'tool_result':
            return {
                'role': 'tool',
                'tool_call_id': component.meta.get('tool_call_id'),
                'content': component.content.content[0].text
            }

    return msg

def _map_to_openai_tool(tool: mcp.server.fastmcp.tools.Tool) -> ChatCompletionToolParam:
    return ChatCompletionToolParam(
        type='function',
        function=FunctionDefinition(
            name=tool.name,
            description=tool.description,
            parameters=tool.parameters
        )
    )

def _map_from_openai_message(message: ChatCompletionMessage) -> Message:
    msg = Message(role=message.role, components=[])

    if message.content:
        msg.components.append(MessageComponent(
            type='message',
            content=message.content
        ))

    if message.reasoning:
        msg.components.append(MessageThinkingComponent(
            type='thinking',
            content=message.reasoning
        ))

    for tool_call in message.tool_calls or []:
        msg.components.append(MessageToolCallComponent(
            meta={'tool_call_id': tool_call.id},
            type='tool_call',
            content=mcp.types.CallToolRequest(
                method='tools/call',
                params=mcp.types.CallToolRequestParams(
                    name=tool_call.function.name,
                    arguments=json.loads(tool_call.function.arguments)
                )
            )
        ))

    return msg

class OpenAICompatibleBackend(Backend):
    def model_post_init(self, ctx):
        self._client = OpenAI(
            api_key=self.config.get('OPENAI_API_KEY'),
            base_url=self.config.get('OPENAI_BASE_URL')
        )

    def generate_response(self, messages, agent_config, tools = [], system_prompt_override: str = "", **kvargs):
        system_prompt = system_prompt_override or agent_config.system_prompt

        mapped_messages=[
            {
                'role': 'system',
                'content': system_prompt
            }
        ]

        for message in messages:
            mapped_messages.append(_map_to_openai_message(message))

        tools_mapped = []
        for tool in tools:
            tools_mapped.append(_map_to_openai_tool(tool))

        response = self._client.chat.completions.create(
            model=agent_config.model_name,
            messages=mapped_messages,
            tools=tools_mapped
        )

        return _map_from_openai_message(response.choices[0].message)



from pydantic import BaseModel, Field
from .agent import Agent
from .types.message import Message, MessageComponent
from typing import Dict, List, Literal, Any, Annotated

import logging
logger = logging.getLogger(__name__)


class FleetDialogBranch(BaseModel):
    meta: Dict[str, Any]

    operating_agent: Agent

    messages: Annotated[List[Message], Field(default_factory=list)]

    def invoke_agent(self, query: str) -> List[Message]:
        self.messages.append(Message(
            role='user',
            components=[MessageComponent(
                type='message',
                content=f"<meta>\n{self.meta}\n</meta><query>{query}</query>"
            )]
        ))

        self.operating_agent._invoke_completion(self.messages, stop_on=['respond_to_agent'], force_tools=True)

        return self.messages



class Fleet(BaseModel):
    agents: Dict[str, Agent] = Field(default_factory=dict)

    _relationships: Dict[str, List[str]]

    _dialog_branches: Dict[str, FleetDialogBranch]

    def model_post_init(self, context):
        self._relationships = {}
        self._dialog_branches = {}
        return super().model_post_init(context)

    def add_agent(self, agent_name: str, agent: Agent, allowed_escalation_agent_names: List[str] =[]):
        self.agents[agent_name] = agent
        self._relationships[agent_name] = allowed_escalation_agent_names

        _agent_name = agent_name
    
        @agent.add_tool
        def respond_to_agent(response: str) -> str:
            """
            Reply to an agent that prompted you
            """
            return response

        if len(allowed_escalation_agent_names) == 0:
            return

        KNOWN_AGENTS_TYPE = Literal[*allowed_escalation_agent_names]

        @agent.add_tool
        def ask_agent(agent_name: KNOWN_AGENTS_TYPE, query: str, context: str):
            """
            Ask a known agent a question
            """
            if agent_name not in allowed_escalation_agent_names:
                return f"ERROR: '{agent_name}' not in {allowed_escalation_agent_names}"
            
            logger.info(f"Escalating to {agent_name}: {query}, {context}")
            
            self._dialog_branches[f"{_agent_name}::{agent_name}"] = FleetDialogBranch(
                meta={
                    "asking_agent": _agent_name
                },
                operating_agent=self.agents.get(agent_name)
            )

            return self._dialog_branches[f"{_agent_name}::{agent_name}"].invoke_agent(f"Question: {query}\nContext: {context}")[-1]

    def invoke_agent(self, agent_name: str, query: str, stop_on: List[str] = []):
        return self.agents.get(agent_name).invoke(query, stop_on = stop_on, force_tools=True)


import ollama

from ..types.backend import Backend
from ..types.message import Message, MessageComponent, MessageThinkingComponent, MessageToolCallComponent

import mcp.types


# HELPERS
def _map_to_ollama_tool(tool) -> ollama.Tool:
    return ollama.Tool(
        type='function',
        function=ollama.Tool.Function(
            name=tool.name,
            description=tool.description,
            parameters=ollama.Tool.Function.Parameters.model_validate(tool.parameters)
        )
    )

def _map_to_ollama_message(message: Message) -> ollama.Message:
    msg_builder = {
        "role": message.role,
        "content": '',
        "thinking": None,
        "tool_calls": []
    }

    for component in message.components:
        if component.type == 'message':
            msg_builder['content'] = component.content
        
        if component.type == 'thinking':
            msg_builder['thinking'] = component.content
        
        if component.type == 'tool_call':
            msg_builder['tool_calls'].append(
                ollama.Message.ToolCall(
                    function=ollama.Message.ToolCall.Function(
                        name=component.content.params.name,
                        arguments=component.content.params.arguments
                    )
                )
            )
        
        if component.type == 'tool_result':
            msg_builder['content'] = f"<{component.meta.get('tool_call_name')}>{component.content.content[0]}</{component.meta.get('tool_call_name')}>"
        
        return ollama.Message.model_validate(msg_builder)


def _map_from_ollama_message(ollama_message: ollama.Message) -> Message:
    msg = Message(role=ollama_message.role, components=[])

    if ollama_message.thinking:
        msg.components.append(MessageThinkingComponent(
            type='thinking',
            content=ollama_message.thinking
        ))

    if ollama_message.content:
        msg.components.append(MessageComponent(
            type='message',
            content=ollama_message.content
        ))

    for tool_call in ollama_message.tool_calls or []:
        msg.components.append(MessageToolCallComponent(
            type='tool_call',
            content=mcp.types.CallToolRequest(
                method='tools/call',
                params=mcp.types.CallToolRequestParams(
                    name=tool_call.function.name,
                    arguments=tool_call.function.arguments
                )
            )
        ))

    return msg



class OllamaBackend(Backend):
    def model_post_init(self, ctx):
        self._client = ollama.Client(
            host="127.0.0.1:11434"
        )

    def generate_response(self, messages, agent_config, tools = [], system_prompt_override: str = "", **kvargs):
        system_prompt = system_prompt_override or agent_config.system_prompt

        mapped_messages=[
            {
                'role': 'system',
                'content': system_prompt
            }
        ]

        for message in messages:
            mapped_messages.append(_map_to_ollama_message(message))

        tools_mapped = []
        for tool in tools:
            tools_mapped.append(_map_to_ollama_tool(tool))

        ollama_response = self._client.chat(
            model=agent_config.model_name,
            messages=mapped_messages,
            tools=tools_mapped,
            think=agent_config.do_thinking
        )

        return _map_from_ollama_message(ollama_response.message)
